from __future__ import annotations

import os
import random
from datetime import datetime
from typing import Any, Optional

from openpyxl import Workbook
from openpyxl.styles import Font, Color, PatternFill, Alignment, Side, Border
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.dimensions import DimensionHolder, ColumnDimension
from openpyxl.worksheet.table import Table
from openpyxl.worksheet.worksheet import Worksheet

from leeger.model.filter import AllTimeFilters, YearFilters
from leeger.model.league import Year, League
from leeger.util.excel_helper import (
    allTimeTeamsStatSheet,
    yearMatchupsStatSheet,
    allTimeMatchupsStatSheet,
)
from leeger.util.navigator.YearNavigator import YearNavigator
from leeger.util.stat_sheet import yearStatSheet, leagueStatSheet


def leagueToExcel(
    league: League, filePath: Optional[str] = None, overwrite: bool = False, **kwargs
) -> Workbook:
    """
    Saves the given League object to an Excel file.
    """

    if league is None:
        raise ValueError("'league' has not been set.")
    if filePath is not None:
        if os.path.exists(filePath) and not overwrite:
            raise FileExistsError(
                f"Cannot create file at path: '{filePath}' because there is already a file there."
            )
        else:
            try:
                os.remove(filePath)
            except FileNotFoundError:
                # we don't care if this doesn't exist
                pass

    # get owner names
    # and
    # make sure we have the same color for owners and their teams across sheets
    ownerIdToSeedMap = dict()
    ownerIds = list()
    ownerIdToNameMap = dict()
    for owner in league.owners:
        ownerIdToSeedMap[owner.id] = f"{owner.id}{datetime.now().date()}"
        ownerIds.append(owner.id)
        ownerIdToNameMap[owner.id] = owner.name
    ownerIdToColorMap = dict()
    for ownerId, seed in ownerIdToSeedMap.items():
        ownerIdToColorMap[ownerId] = _getRandomColor(0.5, seed)

    # gather info for all time teams to be used later
    allTimeTeamIds: list[str] = list()
    teamIdToColorMap = dict()
    workbook = None
    for year in league.years:
        for team in year.teams:
            allTimeTeamIds.append(team.id)
            teamIdToColorMap[team.id] = ownerIdToColorMap[team.ownerId]

        # add a sheet to the Excel document for this year
        workbook = _yearToExcel(year, workbook, **kwargs.copy())

    # add All-Time teams stats sheet
    # figure out index to put this sheet into
    # we want the sheets to be ordered: oldest year -> newest year -> all time teams -> all time matchups -> all time owners
    index = len(workbook.sheetnames)
    workbook.create_sheet("All Time Teams", index=index)
    worksheet = workbook["All Time Teams"]

    allTimeTeamsStatSheet_ = allTimeTeamsStatSheet(league, **kwargs.copy())

    allTimeFilters = AllTimeFilters.preferredOrderWithTitle(league, **kwargs.copy())

    # see if we have a division column
    # if we do, shift the frozen panes 1 column to the right
    freezePanes = "D2"
    for title, _ in allTimeTeamsStatSheet_:
        if title == "Division":
            freezePanes = "E2"

    _populateWorksheet(
        worksheet=worksheet,
        workbook=workbook,
        displayName="AllTimeTeamStats",
        titlesAndStatDicts=allTimeTeamsStatSheet_,
        entityIds=allTimeTeamIds,
        entityIdToColorMap=teamIdToColorMap,
        legendKeyValues=allTimeFilters,
        freezePanes=freezePanes,
        saveToFilepath=filePath,
    )

    # add All-Time matchups stats sheet
    # figure out index to put this sheet into
    # we want the sheets to be ordered: oldest year -> newest year -> all time teams -> all time matchups -> all time owners
    index = len(workbook.sheetnames)
    workbook.create_sheet("All Time Matchups", index=index)
    worksheet = workbook["All Time Matchups"]

    (allTimeMatchupsStatSheet_, modifiedMatchupIdToOwnerIdMap) = allTimeMatchupsStatSheet(
        league, **kwargs.copy()
    )

    modifiedMatchupIdToColorMap: dict = dict()
    for modifiedMatchupId, ownerId in modifiedMatchupIdToOwnerIdMap.items():
        modifiedMatchupIdToColorMap[modifiedMatchupId] = ownerIdToColorMap[ownerId]

    allTimeFilters = AllTimeFilters.preferredOrderWithTitle(league, **kwargs.copy())
    _populateWorksheet(
        worksheet=worksheet,
        workbook=workbook,
        displayName="AllTimeMatchups",
        titlesAndStatDicts=allTimeMatchupsStatSheet_,
        entityIds=list(modifiedMatchupIdToOwnerIdMap.keys()),
        entityIdToColorMap=modifiedMatchupIdToColorMap,
        legendKeyValues=allTimeFilters,
        freezePanes="C2",
        saveToFilepath=filePath,
        boldColumnNumbers=[1, 2],
    )

    # add All-Time owner stats sheet
    # figure out index to put this sheet into
    # we want the sheets to be ordered: oldest year -> newest year -> all time teams -> all time matchups -> all time owners
    index = len(workbook.sheetnames)
    workbook.create_sheet("All Time Owners", index=index)
    worksheet = workbook["All Time Owners"]

    allTimeOwnerStatsWithTitles = leagueStatSheet(league, **kwargs.copy()).preferredOrderWithTitle()
    allTimeOwnerStatsWithTitles.insert(0, ("Owner", ownerIdToNameMap))

    return _populateWorksheet(
        worksheet=worksheet,
        workbook=workbook,
        displayName="AllTimeOwnerStats",
        titlesAndStatDicts=allTimeOwnerStatsWithTitles,
        entityIds=ownerIds,
        entityIdToColorMap=ownerIdToColorMap,
        legendKeyValues=allTimeFilters,
        freezePanes="B2",
        saveToFilepath=filePath,
    )


def _yearToExcel(year: Year, workbook: Optional[Workbook] = None, **kwargs) -> Workbook:
    """
    Saves the given Year object to an Excel File.
    If the given Excel file exists already, add a worksheet to it with this year.
    If the given Excel file does not exist, create a new workbook and add a worksheet to it with this year.
    """

    if workbook is None:
        # create workbook and sheet
        workbook = Workbook()
        workbook.create_sheet(f"{year.yearNumber} Teams", index=0)
        workbook.create_sheet(f"{year.yearNumber} Matchups", index=1)
        # remove default sheet
        del workbook["Sheet"]
    else:
        # figure out index to put this sheet into
        # we want the years as sheets in order from oldest -> newest year
        index = 0
        for sheetname in workbook.sheetnames:
            if year.yearNumber > int(sheetname[:5]):
                index += 1
        workbook.create_sheet(f"{year.yearNumber} Teams", index=index)
        workbook.create_sheet(f"{year.yearNumber} Matchups", index=index + 1)
    worksheet = workbook[f"{year.yearNumber} Teams"]

    # get team names
    # and
    # make sure we have the same color for owners and their teams across sheets
    ownerIdToSeedMap = dict()
    teamIdToNameMap = dict()
    teamIdToDivisionNameMap = dict()
    for team in year.teams:
        ownerIdToSeedMap[team.ownerId] = f"{team.ownerId}{datetime.now().date()}"
        teamIdToNameMap[team.id] = team.name
        if team.divisionId:
            teamIdToDivisionNameMap[team.id] = YearNavigator.getDivisionById(
                year, team.divisionId
            ).name
    ownerIdToColorMap = dict()
    for ownerId, seed in ownerIdToSeedMap.items():
        ownerIdToColorMap[ownerId] = _getRandomColor(0.5, seed)

    teamIdToColorMap = dict()
    teamIds = list()
    for team in year.teams:
        teamIdToColorMap[team.id] = ownerIdToColorMap[team.ownerId]
        teamIds.append(team.id)

    yearFilters = YearFilters.preferredOrderWithTitle(year, **kwargs.copy())

    yearStatsWithTitles = yearStatSheet(year, **kwargs.copy()).preferredOrderWithTitle()
    yearStatsWithTitles.insert(0, ("Team", teamIdToNameMap))

    freezePanes = "B2"
    if len(year.divisions) > 0:
        yearStatsWithTitles.insert(1, ("Division", teamIdToDivisionNameMap))
        freezePanes = "C2"
    # save Year teams to Excel sheet
    _populateWorksheet(
        worksheet=worksheet,
        workbook=workbook,
        displayName=f"Teams{year.yearNumber}",
        titlesAndStatDicts=yearStatsWithTitles,
        entityIds=teamIds,
        entityIdToColorMap=teamIdToColorMap,
        legendKeyValues=yearFilters,
        freezePanes=freezePanes,
    )

    (yearMatchupsWithTitles, modifiedMatchupIdToOwnerIdMap) = yearMatchupsStatSheet(
        year, **kwargs.copy()
    )
    modifiedMatchupIdToColorMap: dict = dict()
    for modifiedMatchupId, ownerId in modifiedMatchupIdToOwnerIdMap.items():
        modifiedMatchupIdToColorMap[modifiedMatchupId] = ownerIdToColorMap[ownerId]
    worksheet = workbook[f"{year.yearNumber} Matchups"]
    # save Year matchups to Excel sheet
    _populateWorksheet(
        worksheet=worksheet,
        workbook=workbook,
        displayName=f"Matchups{year.yearNumber}",
        titlesAndStatDicts=yearMatchupsWithTitles,
        entityIds=list(modifiedMatchupIdToOwnerIdMap.keys()),
        entityIdToColorMap=modifiedMatchupIdToColorMap,
        legendKeyValues=yearFilters,
        freezePanes="B2",
    )
    return workbook


def _getRandomColor(tint: float = 0, seed: str = None) -> Color:
    """
    Used to get a random row color.
    """
    if seed:
        random.seed(seed)
    r = lambda: random.randint(0, 255)
    hexCode = "%02X%02X%02X" % (r(), r(), r())
    return Color(rgb=hexCode, tint=tint)


def _populateWorksheet(
    *,
    worksheet: Worksheet,
    workbook: Workbook,
    saveToFilepath: Optional[str] = None,
    displayName: str,
    titlesAndStatDicts: list[tuple[str, dict]],
    entityIds: list[str],
    entityIdToColorMap: dict[str, Color],
    legendKeyValues: list[tuple[str, Any]],
    freezePanes: str,
    boldColumnNumbers: Optional[list] = None,
) -> Workbook:
    """
    boldColumnNumbers is a list of column numbers whose values will be bolded.
        - Column numbers will be 1-indexed.
        - Default is [1]
    """
    boldColumnNumbers = [1] if boldColumnNumbers is None else boldColumnNumbers
    ####################
    # Styles for table #
    ####################

    # fonts
    HEADER_COLUMN_FONT = Font(size=12, bold=True)
    ENTITY_NAME_FONT = Font(size=11, bold=True)

    # colors
    BLACK = Color(rgb="000000")
    LIGHT_GRAY = Color(rgb="B8B8B8")
    MEDIUM_GRAY = Color(rgb="828282")

    # fills
    HEADER_FILL = PatternFill(patternType="solid", fgColor=LIGHT_GRAY)

    #################
    # Fill in table #
    #################

    # add all stats
    for rowNumber, entityId in enumerate(entityIds):
        rowFill = PatternFill(patternType="solid", fgColor=entityIdToColorMap[entityId])
        for columnNumber, (title, statDict) in enumerate(titlesAndStatDicts):
            char = get_column_letter(columnNumber + 1)
            if rowNumber == 1:
                # add stat header
                cell = f"{char}{rowNumber}"
                worksheet[cell] = title
                worksheet[cell].font = HEADER_COLUMN_FONT
                worksheet[cell].fill = HEADER_FILL
                worksheet[cell].alignment = Alignment(horizontal="center")
            # add stat value
            cell = f"{char}{rowNumber + 2}"
            if entityId in statDict:
                worksheet[cell] = statDict[entityId]
            else:
                worksheet[cell] = "N/A"
            worksheet[cell].fill = rowFill
            if columnNumber + 1 in boldColumnNumbers:
                worksheet[cell].font = ENTITY_NAME_FONT

    # put stats into table
    table = Table(
        displayName=displayName,
        ref="A1:" + get_column_letter(worksheet.max_column) + str(len(entityIds) + 1),
    )
    worksheet.add_table(table)
    # freeze owner name column and header row
    worksheet.freeze_panes = freezePanes

    # add legend for filters
    # define border to go around legend
    topSideSolid = Side(border_style="thick", color=BLACK)
    bottomSideSolid = Side(border_style="thick", color=BLACK)
    bottomSideThin = Side(border_style="thin", color=BLACK)
    leftSideSolid = Side(border_style="thick", color=BLACK)
    rightSideSolid = Side(border_style="thick", color=BLACK)
    # legend formatting
    legendCellAlignment = Alignment(horizontal="center")

    legendRowNumber = len(entityIds) + 4
    legendColLetter = "A"
    titleCell = f"A{legendRowNumber}"
    worksheet[titleCell] = "Filters Applied"
    worksheet[titleCell].fill = PatternFill(patternType="solid", fgColor=MEDIUM_GRAY)
    worksheet[titleCell].font = Font(bold=True)
    border = Border(
        left=leftSideSolid, right=rightSideSolid, top=topSideSolid, bottom=bottomSideThin
    )
    worksheet[titleCell].border = border
    worksheet[titleCell].alignment = legendCellAlignment

    for kwarg_title, kwarg_value in legendKeyValues:
        legendRowNumber += 1
        cell = f"{legendColLetter}{legendRowNumber}"
        worksheet[cell] = f"{kwarg_title}: {kwarg_value}"
        worksheet[cell].fill = PatternFill(patternType="solid", fgColor=MEDIUM_GRAY)
        worksheet[cell].alignment = legendCellAlignment
        border = Border(left=leftSideSolid, right=rightSideSolid)
        worksheet[cell].border = border

    # add border to last cell
    border = Border(left=leftSideSolid, right=rightSideSolid, bottom=bottomSideSolid)
    worksheet[f"{legendColLetter}{legendRowNumber}"].border = border

    # set column widths
    TITLE_MULTIPLIER = 1.2
    NAME_MULTIPLIER = 1.0
    DATA_MULTIPLIER = 0.8
    dim_holder = DimensionHolder(worksheet=worksheet)

    for columnNumber in range(worksheet.min_column, worksheet.max_column + 1):
        # figure out the width we want this column to be
        # first column has entity names, so use a different multiplier for them
        isNameColumn = columnNumber == 1
        maxWidth = 0
        for rowNumber, cell in enumerate(worksheet[get_column_letter(columnNumber)]):
            if cell.value:
                # count title/name cell characters as more than a data cell
                if rowNumber == 0:
                    multiplier = TITLE_MULTIPLIER
                elif isNameColumn:
                    multiplier = NAME_MULTIPLIER
                else:
                    multiplier = DATA_MULTIPLIER
                maxWidth = max((maxWidth, len(str(cell.value)) * multiplier))
        dim_holder[get_column_letter(columnNumber)] = ColumnDimension(
            worksheet, min=columnNumber, max=columnNumber, width=maxWidth + 7
        )

    worksheet.column_dimensions = dim_holder

    # save
    if saveToFilepath is not None:
        workbook.save(saveToFilepath)
    return workbook



from leeger.exception.InvalidDivisionFormatException import InvalidDivisionFormatException
from leeger.model.league.Division import Division


def runAllChecks(division: Division) -> None:
    """
    Runs all checks on the given Division.
    """
    checkAllTypes(division)


def checkAllTypes(division: Division) -> None:
    """
    Checks all types that are within the Owner object.
    """
    if not isinstance(division.name, str):
        raise InvalidDivisionFormatException("name must be type 'str'.")



from typing import Optional
from pymfl.api import CommonLeagueInfoAPIClient
from pymfl.api.config import APIConfig

from leeger.enum.MatchupType import MatchupType
from leeger.league_loader.LeagueLoader import LeagueLoader
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.validate import leagueValidation
from leeger.model.league.Division import Division


class MyFantasyLeagueLeagueLoader(LeagueLoader):
    """
    Responsible for loading a League from MyFantasyLeague.
    http://home.myfantasyleague.com/
    """

    def __init__(
        self,
        leagueId: str,
        years: list[int],
        *,
        mflUsername: str,
        mflPassword: str,
        mflUserAgentName: str,
        ownerNamesAndAliases: Optional[dict[str, list[str]]] = None,
        leagueName: Optional[str] = None,
    ):
        super().__init__(
            leagueId, years, ownerNamesAndAliases=ownerNamesAndAliases, leagueName=leagueName
        )

        self.__mflUsername = mflUsername
        self.__mflPassword = mflPassword
        self.__mflUserAgentName = mflUserAgentName

        self.__mflLeagueIdToYearMap: dict[str, int] = dict()
        self.__mflFranchiseIdToOwnerMap: dict[str, Owner] = dict()
        self.__mflFranchiseIdToTeamMap: dict[int, Team] = dict()
        self.__mflDivisionIdToDivisionMap: dict[
            str, Division
        ] = dict()  # holds the division info for ONLY the current year

    def __getAllLeagues(self) -> list[dict]:
        mflLeagues: list[dict] = list()

        for year in self._years:
            APIConfig.add_config_for_year_and_league_id(
                year=year,
                league_id=self._leagueId,
                username=self.__mflUsername,
                password=self.__mflPassword,
                user_agent_name=self.__mflUserAgentName,
            )

            mflLeague = CommonLeagueInfoAPIClient.get_league(year=year, league_id=self._leagueId)[
                "league"
            ]
            self.__mflLeagueIdToYearMap[mflLeague["id"]] = year
            mflLeagues.append(mflLeague)
        self._validateRetrievedLeagues(mflLeagues)
        return mflLeagues

    def getOwnerNames(self) -> dict[int, list[str]]:
        yearToOwnerNamesMap: dict[int, list[str]] = dict()
        mflLeagues = self.__getAllLeagues()
        for mflLeague in mflLeagues:
            yearNumber = self.__mflLeagueIdToYearMap[mflLeague["id"]]
            yearToOwnerNamesMap[yearNumber] = list()
            for franchise in mflLeague["franchises"]["franchise"]:
                ownerName = franchise["owner_name"]
                yearToOwnerNamesMap[yearNumber].append(ownerName)
        return yearToOwnerNamesMap

    def loadLeague(self, validate: bool = True) -> League:
        mflLeagues = self.__getAllLeagues()
        league = self.__buildLeague(mflLeagues)
        if validate:
            # validate new league
            leagueValidation.runAllChecks(league)
        self._warnForUnusedOwnerNames(league)
        return league

    def __buildLeague(self, mflLeagues: list[dict]) -> League:
        years = list()
        self.__loadOwners(mflLeagues)
        owners = list(self.__mflFranchiseIdToOwnerMap.values())
        for mflLeague in mflLeagues:
            # save league name for each year
            self._leagueNameByYear[self.__mflLeagueIdToYearMap[mflLeague["id"]]] = mflLeague["name"]
            years.append(self.__buildYear(mflLeague))
        return League(name=self._getLeagueName(), owners=owners, years=self._getValidYears(years))

    def __buildYear(self, mflLeague: dict) -> Year:
        # save division info
        for division in mflLeague["divisions"]["division"]:
            self.__mflDivisionIdToDivisionMap[division["id"]] = Division(name=division["name"])
        yearNumber = self.__mflLeagueIdToYearMap[mflLeague["id"]]
        teams = self.__buildTeams(mflLeague)
        weeks = self.__buildWeeks(mflLeague)
        # TODO: see if there are cases where MFL leagues do NOT have divisions
        year = Year(
            yearNumber=yearNumber,
            teams=teams,
            weeks=weeks,
            divisions=list(self.__mflDivisionIdToDivisionMap.values()),
        )
        # clear division info
        self.__mflDivisionIdToDivisionMap = dict()
        return year

    def __buildWeeks(self, mflLeague: dict) -> list[Week]:
        yearNumber = self.__mflLeagueIdToYearMap[mflLeague["id"]]
        weeks = list()
        schedule: dict = CommonLeagueInfoAPIClient.get_schedule(
            year=yearNumber, league_id=mflLeague["id"]
        )["schedule"]
        # get playoff brackets
        playoffBracket: dict = CommonLeagueInfoAPIClient.get_playoff_bracket(
            year=yearNumber, league_id=mflLeague["id"], bracket_id="1"
        )["playoffBracket"]

        # we will assume that the "true" playoff bracket (i.e. the bracket where the winner of it is the league champion)
        # will always be the playoff bracket with id "1".
        # if this changes or is not the case, this will need to be refactored to reflect that.
        # the reason we only care about this bracket (and not the other ones) is because we only need to know who won the championship from this info.
        # we can get all playoff matchup data from the regular schedule

        # if there is only 1 object in the playoffRound field, it is a dict, otherwise it is a list
        playoffWeeks = list()
        if isinstance(playoffBracket["playoffRound"], dict):
            # only 1 playoff round
            playoffWeeks.append(playoffBracket["playoffRound"])
        else:
            # multiple playoff rounds
            for playoffBracketInfo in playoffBracket["playoffRound"]:
                playoffWeeks.append(playoffBracketInfo)

        playoffWeekNumbers = [int(playoffWeek["week"]) for playoffWeek in playoffWeeks]

        playoffsStarted = False
        for week in schedule["weeklySchedule"]:
            weekNumber = int(week["week"])
            if weekNumber > int(mflLeague["lastRegularSeasonWeek"]):
                playoffsStarted = True
            # get each teams matchup for that week
            matchups = list()
            for matchup in week["matchup"]:
                teamAMFLFranchiseId = matchup["franchise"][0]["id"]
                teamAId = self.__mflFranchiseIdToTeamMap[teamAMFLFranchiseId].id
                teamAScore = float(matchup["franchise"][0]["score"])
                teamAHasTiebreaker = matchup["franchise"][0]["result"] == "W"

                teamBMFLFranchiseId = matchup["franchise"][1]["id"]
                teamBId = self.__mflFranchiseIdToTeamMap[teamBMFLFranchiseId].id
                teamBScore = float(matchup["franchise"][1]["score"])
                teamBHasTiebreaker = matchup["franchise"][1]["result"] == "W"

                isChampionshipMatchup = False

                matchupType = MatchupType.REGULAR_SEASON
                if playoffsStarted:
                    # this is a playoff matchup or a championship matchup
                    # NOTE: If there are MFL leagues that have matchups during playoff weeks that are
                    # NOTE: NOT playoff matchups, this logic will need to be changed.
                    for playoffWeek in playoffWeeks:
                        currentIsChampionshipMatchup = self.__isChampionshipMatchup(
                            playoffWeek=playoffWeek,
                            playoffWeekNumbers=playoffWeekNumbers,
                            weekNumber=weekNumber,
                            teamAMFLFranchiseId=teamAMFLFranchiseId,
                            teamBMFLFranchiseId=teamBMFLFranchiseId,
                        )
                        isChampionshipMatchup = (
                            isChampionshipMatchup or currentIsChampionshipMatchup
                        )
                    matchupType = MatchupType.PLAYOFF
                    if isChampionshipMatchup:
                        matchupType = MatchupType.CHAMPIONSHIP

                matchups.append(
                    Matchup(
                        teamAId=teamAId,
                        teamBId=teamBId,
                        teamAScore=teamAScore,
                        teamBScore=teamBScore,
                        teamAHasTiebreaker=teamAHasTiebreaker,
                        teamBHasTiebreaker=teamBHasTiebreaker,
                        matchupType=matchupType,
                    )
                )
            if len(matchups) > 0:
                weeks.append(Week(weekNumber=weekNumber, matchups=matchups))

        return weeks

    @staticmethod
    def __isChampionshipMatchup(
        *,
        playoffWeek: dict,
        playoffWeekNumbers: list[int],
        weekNumber: int,
        teamAMFLFranchiseId: int,
        teamBMFLFranchiseId: int,
    ) -> tuple[bool, bool]:
        # helper method
        def isValid(*, pGame: dict, aId: int, bId: int) -> bool:
            return (
                aId == pGame["away"]["franchise_id"] or bId == pGame["away"]["franchise_id"]
            ) and (aId == pGame["home"]["franchise_id"] or bId == pGame["home"]["franchise_id"])

        isChampionshipMatchup = False

        if weekNumber == max(playoffWeekNumbers):
            # this is the last week in the bracket, the championship week
            # if there is only 1 object in the playoffGame field, it is a dict, otherwise it is a list
            if isinstance(playoffWeek["playoffGame"], dict):
                # only 1 game
                # check if this matchup is the championship game
                isChampionshipMatchup = isValid(
                    pGame=playoffWeek["playoffGame"],
                    aId=teamAMFLFranchiseId,
                    bId=teamBMFLFranchiseId,
                )
            else:
                # multiple games
                # check if this matchup is the championship game
                for playoffGame in playoffWeek["playoffGame"]:
                    isChampionshipMatchup = isChampionshipMatchup or isValid(
                        pGame=playoffGame, aId=teamAMFLFranchiseId, bId=teamBMFLFranchiseId
                    )
        return isChampionshipMatchup

    def __buildTeams(self, mflLeague: dict) -> list[Team]:
        teams = list()
        for franchise in mflLeague["franchises"]["franchise"]:
            # TODO: see if there are cases where MFL leagues do NOT have divisions
            divisionId = self.__mflDivisionIdToDivisionMap[franchise["division"]].id
            owner = self.__mflFranchiseIdToOwnerMap[franchise["id"]]
            team = Team(name=franchise["name"], ownerId=owner.id, divisionId=divisionId)
            self.__mflFranchiseIdToTeamMap[franchise["id"]] = team
            teams.append(team)
        return teams

    def __loadOwners(self, mflLeagues: list[dict]) -> None:
        for mflLeague in mflLeagues:
            for franchise in mflLeague["franchises"]["franchise"]:
                ownerName = franchise["owner_name"]
                # get general owner name if there is one
                generalOwnerName = self._getGeneralOwnerNameFromGivenOwnerName(ownerName)
                ownerName = generalOwnerName if generalOwnerName is not None else ownerName
                self.__mflFranchiseIdToOwnerMap[franchise["id"]] = Owner(name=ownerName)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from leeger.enum.MatchupType import MatchupType
from leeger.exception import DoesNotExistException
from leeger.exception.InvalidMatchupFormatException import InvalidMatchupFormatException
from leeger.model.abstract.EqualityCheck import EqualityCheck
from leeger.model.abstract.UniqueId import UniqueId
from leeger.model.league_helper.Performance import Performance
from leeger.util.CustomLogger import CustomLogger
from leeger.util.JSONDeserializable import JSONDeserializable
from leeger.util.JSONSerializable import JSONSerializable
from leeger.util.equality import modelEquals


@dataclass(kw_only=True, eq=False)
class Matchup(UniqueId, EqualityCheck, JSONSerializable, JSONDeserializable):
    __LOGGER = CustomLogger.getLogger()
    teamAId: str
    teamBId: str
    teamAScore: float | int
    teamBScore: float | int
    matchupType: MatchupType = MatchupType.REGULAR_SEASON
    teamAHasTiebreaker: Optional[bool] = False
    teamBHasTiebreaker: Optional[bool] = False
    multiWeekMatchupId: Optional[
        str
    ] = None  # This is used to link matchups that span over multiple weeks

    def __post_init__(self):
        # Team A and Team B cannot both have the tiebreaker
        if self.teamAHasTiebreaker is True and self.teamBHasTiebreaker is True:
            raise InvalidMatchupFormatException(
                "Team A and Team B cannot both have the tiebreaker."
            )

        if self.teamAHasTiebreaker is None:
            self.teamAHasTiebreaker = False
        if self.teamBHasTiebreaker is None:
            self.teamBHasTiebreaker = False

    def equals(
        self,
        otherMatchup: Matchup,
        *,
        ignoreIds: bool = False,
        ignoreBaseIds: bool = False,
        logDifferences: bool = False,
    ) -> bool:
        """
        Checks if *this* Matchup is the same as the given Matchup.
        """

        return modelEquals(
            objA=self,
            objB=otherMatchup,
            baseFields={
                "teamAScore",
                "teamBScore",
                "matchupType",
                "teamAHasTiebreaker",
                "teamBHasTiebreaker",
            },
            idFields={"teamAId", "teamBId", "multiWeekMatchupId"},
            parentKey="Matchup",
            ignoreIdFields=ignoreIds,
            ignoreBaseIdField=ignoreBaseIds,
            logDifferences=logDifferences,
        )

    def __eq__(self, otherMatchup: Matchup) -> bool:
        self.__LOGGER.info("Use .equals() for more options when comparing Matchup instances.")
        return self.equals(otherMatchup=otherMatchup)

    def splitToPerformances(self) -> tuple[Performance, Performance]:
        """
        Splits this Matchup into 2 Performances.
        """
        performanceA = Performance(
            teamId=self.teamAId,
            teamScore=self.teamAScore,
            matchupType=self.matchupType,
            multiWeekMatchupId=self.multiWeekMatchupId,
        )
        performanceB = Performance(
            teamId=self.teamBId,
            teamScore=self.teamBScore,
            matchupType=self.matchupType,
            multiWeekMatchupId=self.multiWeekMatchupId,
        )
        return performanceA, performanceB

    def getPerformanceForTeamId(self, teamId: str) -> Performance:
        performances = self.splitToPerformances()
        for performance in performances:
            if performance.teamId == teamId:
                return performance
        raise DoesNotExistException(f"Matchup does not have a team with ID '{teamId}'.")

    def toJson(self) -> dict:
        return {
            "id": self.id,
            "teamAId": self.teamAId,
            "teamBId": self.teamBId,
            "teamAScore": self.teamAScore,
            "teamBScore": self.teamBScore,
            "matchupType": self.matchupType.name,
            "teamAHasTiebreaker": self.teamAHasTiebreaker,
            "teamBHasTiebreaker": self.teamBHasTiebreaker,
            "multiWeekMatchupId": self.multiWeekMatchupId,
        }

    @staticmethod
    def fromJson(d: dict) -> Matchup:
        matchup = Matchup(
            teamAId=d["teamAId"],
            teamBId=d["teamBId"],
            teamAScore=d["teamAScore"],
            teamBScore=d["teamBScore"],
            matchupType=MatchupType.fromStr(d["matchupType"]),
            teamAHasTiebreaker=d.get("teamAHasTiebreaker"),
            teamBHasTiebreaker=d.get("teamBHasTiebreaker"),
            multiWeekMatchupId=d.get("multiWeekMatchupId"),
        )
        matchup.id = d["id"]
        return matchup



import unittest

from leeger.enum.MatchupType import MatchupType
from leeger.exception.InvalidWeekFormatException import InvalidWeekFormatException
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Week import Week
from leeger.validate import weekValidation


class TestWeekValidation(unittest.TestCase):
    def test_checkWeekHasAtLeastOneMatchup_weekDoesntHaveAtLeastOneMatchup_raisesException(self):
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkWeekHasAtLeastOneMatchup(Week(weekNumber=1, matchups=list()))
        self.assertEqual("Week 1 must have at least 1 matchup.", str(context.exception))

    def test_checkForDuplicateMatchups_duplicateMatchupInstances_raisesException(self):
        matchup = Matchup(teamAId="a", teamBId="b", teamAScore=1, teamBScore=2)
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkForDuplicateMatchups(
                Week(weekNumber=1, matchups=[matchup, matchup])
            )
        self.assertEqual("Matchups must all be unique instances.", str(context.exception))

    def test_checkWeekDoesNotHaveMoreThanOneChampionshipMatchup_multipleChampionshipMatchups_raisesException(
        self,
    ):
        matchup1 = Matchup(
            teamAId="a",
            teamBId="b",
            teamAScore=1,
            teamBScore=2,
            matchupType=MatchupType.CHAMPIONSHIP,
        )
        matchup2 = Matchup(
            teamAId="a",
            teamBId="b",
            teamAScore=1,
            teamBScore=2,
            matchupType=MatchupType.CHAMPIONSHIP,
        )
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkWeekDoesNotHaveMoreThanOneChampionshipMatchup(
                Week(weekNumber=1, matchups=[matchup1, matchup2])
            )
        self.assertEqual(
            "Week 1 has 2 championship matchups. Maximum is 1.", str(context.exception)
        )

    def test_checkWeekHasMatchupsWithNoDuplicateTeamIds_teamHasMultipleMatchupsInWeek_raisesException(
        self,
    ):
        matchup1 = Matchup(teamAId="a", teamBId="b", teamAScore=1, teamBScore=2)
        matchup2 = Matchup(teamAId="c", teamBId="b", teamAScore=1, teamBScore=2)
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkWeekHasMatchupsWithNoDuplicateTeamIds(
                Week(weekNumber=1, matchups=[matchup1, matchup2])
            )
        self.assertEqual("Week 1 has matchups with duplicate team IDs.", str(context.exception))

    def test_checkWeekWithPlayoffOrChampionshipMatchupDoesNotHaveRegularSeasonMatchup_weekHasPlayoffMatchupAndRegularSeasonMatchup_raisesException(
        self,
    ):
        matchup1 = Matchup(
            teamAId="a", teamBId="b", teamAScore=1, teamBScore=2, matchupType=MatchupType.PLAYOFF
        )
        matchup2 = Matchup(
            teamAId="c",
            teamBId="b",
            teamAScore=1,
            teamBScore=2,
            matchupType=MatchupType.REGULAR_SEASON,
        )
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkWeekWithPlayoffOrChampionshipMatchupDoesNotHaveRegularSeasonMatchup(
                Week(weekNumber=1, matchups=[matchup1, matchup2])
            )
        self.assertEqual(
            "Week 1 has regular season matchups and playoff/championship matchups in the same week.",
            str(context.exception),
        )

    def test_checkWeekWithPlayoffOrChampionshipMatchupDoesNotHaveRegularSeasonMatchup_weekHasChampionshipMatchupAndRegularSeasonMatchup_raisesException(
        self,
    ):
        matchup1 = Matchup(
            teamAId="a",
            teamBId="b",
            teamAScore=1,
            teamBScore=2,
            matchupType=MatchupType.CHAMPIONSHIP,
        )
        matchup2 = Matchup(
            teamAId="c",
            teamBId="b",
            teamAScore=1,
            teamBScore=2,
            matchupType=MatchupType.REGULAR_SEASON,
        )
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkWeekWithPlayoffOrChampionshipMatchupDoesNotHaveRegularSeasonMatchup(
                Week(weekNumber=1, matchups=[matchup1, matchup2])
            )
        self.assertEqual(
            "Week 1 has regular season matchups and playoff/championship matchups in the same week.",
            str(context.exception),
        )

    def test_checkMultiWeekMatchupsWithSameIdAreOnlyInOneMatchupPerWeek_weekHasSameMultiMatchupIdInMultipleMatchups_raisesException(
        self,
    ):
        matchup1 = Matchup(
            teamAId="a", teamBId="b", teamAScore=1, teamBScore=2, multiWeekMatchupId="1"
        )
        matchup2 = Matchup(
            teamAId="c", teamBId="b", teamAScore=1, teamBScore=2, multiWeekMatchupId="1"
        )
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkMultiWeekMatchupsWithSameIdAreOnlyInOneMatchupPerWeek(
                Week(weekNumber=1, matchups=[matchup1, matchup2])
            )
        self.assertEqual(
            "Week 1 has the multi-week matchup ID '1' in multiple matchups.", str(context.exception)
        )

    """
    TYPE CHECK TESTS
    """

    def test_checkAllTypes_weekNumberIsntTypeInt_raisesException(self):
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkAllTypes(Week(weekNumber=None, matchups=list()))
        self.assertEqual("weekNumber must be type 'int'.", str(context.exception))

    def test_checkAllTypes_weekMatchupsIsntTypeList_raisesException(self):
        # not given a list
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkAllTypes(Week(weekNumber=1, matchups=None))
        self.assertEqual("matchups must be type 'list[Matchup]'.", str(context.exception))

        # given a list of non Matchup
        with self.assertRaises(InvalidWeekFormatException) as context:
            weekValidation.checkAllTypes(Week(weekNumber=1, matchups=["foo"]))
        self.assertEqual("matchups must be type 'list[Matchup]'.", str(context.exception))



import unittest
from unittest import mock
import copy
from leeger.enum.MatchupType import MatchupType

from leeger.league_loader import MyFantasyLeagueLeagueLoader
from leeger.model.league.Division import Division
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year


class TestMyFantasyLeagueLeagueLoader(unittest.TestCase):
    def __addScoreToMockFranchise(
        self, *, mockFranchise: dict, score: int | float, result: str
    ) -> dict:
        dupMockFranchise = copy.deepcopy(mockFranchise)
        dupMockFranchise["score"] = score
        dupMockFranchise["result"] = result
        return dupMockFranchise

    @mock.patch("pymfl.api.config.APIConfig.add_config_for_year_and_league_id")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_league")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_schedule")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_playoff_bracket")
    def test_loadLeague_happyPath(
        self, mockGetPlayoffBracket, mockGetSchedule, mockGetLeague, mockAddConfig
    ):
        """
        NOTE: This also tests the following cases:
            - having 1 playoff round in a year (dict)
            - having multiple playoff rounds in a year (list)
            - having 1 playoff matchup in a playoff week (dict)
            - having multiple playoff matchups in a playoff week (list)
        """
        mockFranchise1_2022 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2022 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2022 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2022 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2022 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2022 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2022 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2022 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2023 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2023 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2023 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2023 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2023 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2023 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2023 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2023 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2024 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2024 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2024 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2024 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2024 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2024 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2024 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2024 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockLeague2022 = {
            "league": {
                "id": 123,
                "name": "Test League 2022",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2022,
                        mockFranchise2_2022,
                        mockFranchise3_2022,
                        mockFranchise4_2022,
                        mockFranchise5_2022,
                        mockFranchise6_2022,
                        mockFranchise7_2022,
                        mockFranchise8_2022,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2022"}, {"id": "2", "name": "d2_2022"}]
                },
            }
        }

        mockLeague2023 = {
            "league": {
                "id": 456,
                "name": "Test League 2023",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2023,
                        mockFranchise2_2023,
                        mockFranchise3_2023,
                        mockFranchise4_2023,
                        mockFranchise5_2023,
                        mockFranchise6_2023,
                        mockFranchise7_2023,
                        mockFranchise8_2023,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2023"}, {"id": "2", "name": "d2_2023"}]
                },
            }
        }

        mockLeague2024 = {
            "league": {
                "id": 789,
                "name": "Test League 2024",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2024,
                        mockFranchise2_2024,
                        mockFranchise3_2024,
                        mockFranchise4_2024,
                        mockFranchise5_2024,
                        mockFranchise6_2024,
                        mockFranchise7_2024,
                        mockFranchise8_2024,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2024"}, {"id": "2", "name": "d2_2024"}]
                },
            }
        }

        mockSchedule2022 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2023 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2023, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2023, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2023, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=93, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=93, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "4",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2024 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockPlayoffSchedule_2022 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2023 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": [
                            {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                            {"away": {"franchise_id": 4}, "home": {"franchise_id": 5}},
                        ],
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 4}},
                    },
                    {
                        "week": 4,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2024 = {
            "playoffBracket": {
                "playoffRound": {
                    "week": 2,
                    "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                }
            }
        }

        mockGetLeague.side_effect = [mockLeague2022, mockLeague2023, mockLeague2024]
        mockGetSchedule.side_effect = [mockSchedule2022, mockSchedule2023, mockSchedule2024]
        mockGetPlayoffBracket.side_effect = [
            mockPlayoffSchedule_2022,
            mockPlayoffSchedule_2023,
            mockPlayoffSchedule_2024,
        ]

        leagueLoader = MyFantasyLeagueLeagueLoader(
            "789",
            [2022, 2023, 2024],
            mflUsername="mflu",
            mflPassword="mflp",
            mflUserAgentName="mfluan",
        )
        league = leagueLoader.loadLeague()

        # expected league

        division1_2022 = Division(name="d1_2022")
        division2_2022 = Division(name="d2_2022")

        division1_2023 = Division(name="d1_2023")
        division2_2023 = Division(name="d2_2023")

        division1_2024 = Division(name="d1_2024")
        division2_2024 = Division(name="d2_2024")

        owner1 = Owner(name="Owner 1")
        owner2 = Owner(name="Owner 2")
        owner3 = Owner(name="Owner 3")
        owner4 = Owner(name="Owner 4")
        owner5 = Owner(name="Owner 5")
        owner6 = Owner(name="Owner 6")
        owner7 = Owner(name="Owner 7")
        owner8 = Owner(name="Owner 8")

        team1_2022 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2022.id)
        team2_2022 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2022.id)
        team3_2022 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2022.id)
        team4_2022 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2022.id)
        team5_2022 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2022.id)
        team6_2022 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2022.id)
        team7_2022 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2022.id)
        team8_2022 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2022.id)

        team1_2023 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2023.id)
        team2_2023 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2023.id)
        team3_2023 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2023.id)
        team4_2023 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2023.id)
        team5_2023 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2023.id)
        team6_2023 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2023.id)
        team7_2023 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2023.id)
        team8_2023 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2023.id)

        team1_2024 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2024.id)
        team2_2024 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2024.id)
        team3_2024 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2024.id)
        team4_2024 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2024.id)
        team5_2024 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2024.id)
        team6_2024 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2024.id)
        team7_2024 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2024.id)
        team8_2024 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2024.id)

        expectedLeague = League(
            name="Test League 2024",
            owners=[owner1, owner2, owner3, owner4, owner5, owner6, owner7, owner8],
            years=[
                Year(
                    yearNumber=2022,
                    teams=[
                        team1_2022,
                        team2_2022,
                        team3_2022,
                        team4_2022,
                        team5_2022,
                        team6_2022,
                        team7_2022,
                        team8_2022,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2022.id,
                                    teamBId=team4_2022.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2022.id,
                                    teamBId=team6_2022.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2022.id,
                                    teamBId=team8_2022.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2022.id,
                                    teamBId=team3_2022.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2022, division2_2022],
                ),
                Year(
                    yearNumber=2023,
                    teams=[
                        team1_2023,
                        team2_2023,
                        team3_2023,
                        team4_2023,
                        team5_2023,
                        team6_2023,
                        team7_2023,
                        team8_2023,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2023.id,
                                    teamBId=team6_2023.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2023.id,
                                    teamBId=team8_2023.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team3_2023.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team4_2023.id,
                                    teamBId=team5_2023.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=4,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2023, division2_2023],
                ),
                Year(
                    yearNumber=2024,
                    teams=[
                        team1_2024,
                        team2_2024,
                        team3_2024,
                        team4_2024,
                        team5_2024,
                        team6_2024,
                        team7_2024,
                        team8_2024,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2024.id,
                                    teamBId=team2_2024.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2024.id,
                                    teamBId=team4_2024.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2024.id,
                                    teamBId=team6_2024.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2024.id,
                                    teamBId=team8_2024.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2024.id,
                                    teamBId=team2_2024.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2024, division2_2024],
                ),
            ],
        )

        self.assertTrue(league.equals(expectedLeague, ignoreBaseIds=True, ignoreIds=True))
        # check multiWeekMatchupIds
        for year in league.years:
            for week in year.weeks:
                for matchup in week.matchups:
                    self.assertIsNone(matchup.multiWeekMatchupId)

    @mock.patch("pymfl.api.config.APIConfig.add_config_for_year_and_league_id")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_league")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_schedule")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_playoff_bracket")
    def test_loadLeague_happyPath_withOwnerNamesAndAliases(
        self, mockGetPlayoffBracket, mockGetSchedule, mockGetLeague, mockAddConfig
    ):
        """
        NOTE: This also tests the following cases:
            - having 1 playoff round in a year (dict)
            - having multiple playoff rounds in a year (list)
            - having 1 playoff matchup in a playoff week (dict)
            - having multiple playoff matchups in a playoff week (list)
        """
        mockFranchise1_2022 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2022 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2022 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2022 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2022 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2022 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2022 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2022 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2023 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2023 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2023 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2023 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2023 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2023 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2023 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2023 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2024 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2024 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2024 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2024 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2024 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2024 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2024 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2024 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockLeague2022 = {
            "league": {
                "id": 123,
                "name": "Test League 2022",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2022,
                        mockFranchise2_2022,
                        mockFranchise3_2022,
                        mockFranchise4_2022,
                        mockFranchise5_2022,
                        mockFranchise6_2022,
                        mockFranchise7_2022,
                        mockFranchise8_2022,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2022"}, {"id": "2", "name": "d2_2022"}]
                },
            }
        }

        mockLeague2023 = {
            "league": {
                "id": 456,
                "name": "Test League 2023",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2023,
                        mockFranchise2_2023,
                        mockFranchise3_2023,
                        mockFranchise4_2023,
                        mockFranchise5_2023,
                        mockFranchise6_2023,
                        mockFranchise7_2023,
                        mockFranchise8_2023,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2023"}, {"id": "2", "name": "d2_2023"}]
                },
            }
        }

        mockLeague2024 = {
            "league": {
                "id": 789,
                "name": "Test League 2024",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2024,
                        mockFranchise2_2024,
                        mockFranchise3_2024,
                        mockFranchise4_2024,
                        mockFranchise5_2024,
                        mockFranchise6_2024,
                        mockFranchise7_2024,
                        mockFranchise8_2024,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2024"}, {"id": "2", "name": "d2_2024"}]
                },
            }
        }

        mockSchedule2022 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2023 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2023, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2023, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2023, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=93, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=93, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "4",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2024 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockPlayoffSchedule_2022 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2023 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": [
                            {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                            {"away": {"franchise_id": 4}, "home": {"franchise_id": 5}},
                        ],
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 4}},
                    },
                    {
                        "week": 4,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2024 = {
            "playoffBracket": {
                "playoffRound": {
                    "week": 2,
                    "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                }
            }
        }

        mockGetLeague.side_effect = [mockLeague2022, mockLeague2023, mockLeague2024]
        mockGetSchedule.side_effect = [mockSchedule2022, mockSchedule2023, mockSchedule2024]
        mockGetPlayoffBracket.side_effect = [
            mockPlayoffSchedule_2022,
            mockPlayoffSchedule_2023,
            mockPlayoffSchedule_2024,
        ]

        leagueLoader = MyFantasyLeagueLeagueLoader(
            "789",
            [2022, 2023, 2024],
            mflUsername="mflu",
            mflPassword="mflp",
            mflUserAgentName="mfluan",
            ownerNamesAndAliases={
                "o1": ["Owner 1"],
                "o2": ["Owner 2"],
                "o3": ["Owner 3"],
                "o4": ["Owner 4"],
                "o5": ["Owner 5"],
                "o6": ["Owner 6"],
                "o7": ["Owner 7"],
                "o8": ["Owner 8"],
            },
        )
        league = leagueLoader.loadLeague()

        # expected league

        division1_2022 = Division(name="d1_2022")
        division2_2022 = Division(name="d2_2022")

        division1_2023 = Division(name="d1_2023")
        division2_2023 = Division(name="d2_2023")

        division1_2024 = Division(name="d1_2024")
        division2_2024 = Division(name="d2_2024")

        owner1 = Owner(name="o1")
        owner2 = Owner(name="o2")
        owner3 = Owner(name="o3")
        owner4 = Owner(name="o4")
        owner5 = Owner(name="o5")
        owner6 = Owner(name="o6")
        owner7 = Owner(name="o7")
        owner8 = Owner(name="o8")

        team1_2022 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2022.id)
        team2_2022 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2022.id)
        team3_2022 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2022.id)
        team4_2022 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2022.id)
        team5_2022 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2022.id)
        team6_2022 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2022.id)
        team7_2022 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2022.id)
        team8_2022 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2022.id)

        team1_2023 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2023.id)
        team2_2023 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2023.id)
        team3_2023 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2023.id)
        team4_2023 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2023.id)
        team5_2023 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2023.id)
        team6_2023 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2023.id)
        team7_2023 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2023.id)
        team8_2023 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2023.id)

        team1_2024 = Team(ownerId=owner1.id, name="Team 1", divisionId=division1_2024.id)
        team2_2024 = Team(ownerId=owner2.id, name="Team 2", divisionId=division1_2024.id)
        team3_2024 = Team(ownerId=owner3.id, name="Team 3", divisionId=division1_2024.id)
        team4_2024 = Team(ownerId=owner4.id, name="Team 4", divisionId=division1_2024.id)
        team5_2024 = Team(ownerId=owner5.id, name="Team 5", divisionId=division2_2024.id)
        team6_2024 = Team(ownerId=owner6.id, name="Team 6", divisionId=division2_2024.id)
        team7_2024 = Team(ownerId=owner7.id, name="Team 7", divisionId=division2_2024.id)
        team8_2024 = Team(ownerId=owner8.id, name="Team 8", divisionId=division2_2024.id)

        expectedLeague = League(
            name="Test League 2024",
            owners=[owner1, owner2, owner3, owner4, owner5, owner6, owner7, owner8],
            years=[
                Year(
                    yearNumber=2022,
                    teams=[
                        team1_2022,
                        team2_2022,
                        team3_2022,
                        team4_2022,
                        team5_2022,
                        team6_2022,
                        team7_2022,
                        team8_2022,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2022.id,
                                    teamBId=team4_2022.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2022.id,
                                    teamBId=team6_2022.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2022.id,
                                    teamBId=team8_2022.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2022.id,
                                    teamBId=team3_2022.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2022.id,
                                    teamBId=team2_2022.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2022, division2_2022],
                ),
                Year(
                    yearNumber=2023,
                    teams=[
                        team1_2023,
                        team2_2023,
                        team3_2023,
                        team4_2023,
                        team5_2023,
                        team6_2023,
                        team7_2023,
                        team8_2023,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2023.id,
                                    teamBId=team6_2023.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2023.id,
                                    teamBId=team8_2023.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team3_2023.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team4_2023.id,
                                    teamBId=team5_2023.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=3,
                            matchups=[
                                Matchup(
                                    teamAId=team2_2023.id,
                                    teamBId=team4_2023.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.PLAYOFF,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                        Week(
                            weekNumber=4,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2023.id,
                                    teamBId=team2_2023.id,
                                    teamAScore=104,
                                    teamBScore=94,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2023, division2_2023],
                ),
                Year(
                    yearNumber=2024,
                    teams=[
                        team1_2024,
                        team2_2024,
                        team3_2024,
                        team4_2024,
                        team5_2024,
                        team6_2024,
                        team7_2024,
                        team8_2024,
                    ],
                    weeks=[
                        Week(
                            weekNumber=1,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2024.id,
                                    teamBId=team2_2024.id,
                                    teamAScore=100,
                                    teamBScore=100,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team3_2024.id,
                                    teamBId=team4_2024.id,
                                    teamAScore=100.1,
                                    teamBScore=90.1,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team5_2024.id,
                                    teamBId=team6_2024.id,
                                    teamAScore=101,
                                    teamBScore=91,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                                Matchup(
                                    teamAId=team7_2024.id,
                                    teamBId=team8_2024.id,
                                    teamAScore=102,
                                    teamBScore=92,
                                    matchupType=MatchupType.REGULAR_SEASON,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                ),
                            ],
                        ),
                        Week(
                            weekNumber=2,
                            matchups=[
                                Matchup(
                                    teamAId=team1_2024.id,
                                    teamBId=team2_2024.id,
                                    teamAScore=103,
                                    teamBScore=93,
                                    matchupType=MatchupType.CHAMPIONSHIP,
                                    teamAHasTiebreaker=True,
                                    teamBHasTiebreaker=False,
                                )
                            ],
                        ),
                    ],
                    yearSettings=None,
                    divisions=[division1_2024, division2_2024],
                ),
            ],
        )

        self.assertTrue(league.equals(expectedLeague, ignoreBaseIds=True, ignoreIds=True))
        # check multiWeekMatchupIds
        for year in league.years:
            for week in year.weeks:
                for matchup in week.matchups:
                    self.assertIsNone(matchup.multiWeekMatchupId)

    @mock.patch("pymfl.api.config.APIConfig.add_config_for_year_and_league_id")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_league")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_schedule")
    @mock.patch("pymfl.api.CommonLeagueInfoAPIClient.get_playoff_bracket")
    def test_loadLeague_withLeagueName(
        self, mockGetPlayoffBracket, mockGetSchedule, mockGetLeague, mockAddConfig
    ):
        mockFranchise1_2022 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2022 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2022 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2022 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2022 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2022 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2022 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2022 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2023 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2023 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2023 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2023 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2023 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2023 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2023 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2023 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockFranchise1_2024 = {"owner_name": "Owner 1", "name": "Team 1", "id": 1, "division": "1"}
        mockFranchise2_2024 = {"owner_name": "Owner 2", "name": "Team 2", "id": 2, "division": "1"}
        mockFranchise3_2024 = {"owner_name": "Owner 3", "name": "Team 3", "id": 3, "division": "1"}
        mockFranchise4_2024 = {"owner_name": "Owner 4", "name": "Team 4", "id": 4, "division": "1"}
        mockFranchise5_2024 = {"owner_name": "Owner 5", "name": "Team 5", "id": 5, "division": "2"}
        mockFranchise6_2024 = {"owner_name": "Owner 6", "name": "Team 6", "id": 6, "division": "2"}
        mockFranchise7_2024 = {"owner_name": "Owner 7", "name": "Team 7", "id": 7, "division": "2"}
        mockFranchise8_2024 = {"owner_name": "Owner 8", "name": "Team 8", "id": 8, "division": "2"}

        mockLeague2022 = {
            "league": {
                "id": 123,
                "name": "Test League 2022",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2022,
                        mockFranchise2_2022,
                        mockFranchise3_2022,
                        mockFranchise4_2022,
                        mockFranchise5_2022,
                        mockFranchise6_2022,
                        mockFranchise7_2022,
                        mockFranchise8_2022,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2022"}, {"id": "2", "name": "d2_2022"}]
                },
            }
        }

        mockLeague2023 = {
            "league": {
                "id": 456,
                "name": "Test League 2023",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2023,
                        mockFranchise2_2023,
                        mockFranchise3_2023,
                        mockFranchise4_2023,
                        mockFranchise5_2023,
                        mockFranchise6_2023,
                        mockFranchise7_2023,
                        mockFranchise8_2023,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2023"}, {"id": "2", "name": "d2_2023"}]
                },
            }
        }

        mockLeague2024 = {
            "league": {
                "id": 789,
                "name": "Test League 2024",
                "lastRegularSeasonWeek": "1",
                "franchises": {
                    "franchise": [
                        mockFranchise1_2024,
                        mockFranchise2_2024,
                        mockFranchise3_2024,
                        mockFranchise4_2024,
                        mockFranchise5_2024,
                        mockFranchise6_2024,
                        mockFranchise7_2024,
                        mockFranchise8_2024,
                    ]
                },
                "divisions": {
                    "division": [{"id": "1", "name": "d1_2024"}, {"id": "2", "name": "d2_2024"}]
                },
            }
        }

        mockSchedule2022 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2023 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2023, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2023, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2023, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=93, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2023, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2023, score=93, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "3",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                    {
                        "week": "4",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2023, score=104, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2023, score=94, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockSchedule2024 = {
            "schedule": {
                "weeklySchedule": [
                    {
                        "week": "1",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=100, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=100, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise3_2022, score=100.1, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise4_2022, score=90.1, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise5_2022, score=101, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise6_2022, score=91, result="L"
                                    ),
                                ]
                            },
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise7_2022, score=102, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise8_2022, score=92, result="L"
                                    ),
                                ]
                            },
                        ],
                    },
                    {
                        "week": "2",
                        "matchup": [
                            {
                                "franchise": [
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise1_2022, score=103, result="W"
                                    ),
                                    self.__addScoreToMockFranchise(
                                        mockFranchise=mockFranchise2_2022, score=93, result="L"
                                    ),
                                ]
                            }
                        ],
                    },
                ]
            }
        }

        mockPlayoffSchedule_2022 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2023 = {
            "playoffBracket": {
                "playoffRound": [
                    {
                        "week": 2,
                        "playoffGame": [
                            {"away": {"franchise_id": 2}, "home": {"franchise_id": 3}},
                            {"away": {"franchise_id": 4}, "home": {"franchise_id": 5}},
                        ],
                    },
                    {
                        "week": 3,
                        "playoffGame": {"away": {"franchise_id": 2}, "home": {"franchise_id": 4}},
                    },
                    {
                        "week": 4,
                        "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                    },
                ]
            }
        }

        mockPlayoffSchedule_2024 = {
            "playoffBracket": {
                "playoffRound": {
                    "week": 2,
                    "playoffGame": {"away": {"franchise_id": 1}, "home": {"franchise_id": 2}},
                }
            }
        }

        mockGetLeague.side_effect = [mockLeague2022, mockLeague2023, mockLeague2024]
        mockGetSchedule.side_effect = [mockSchedule2022, mockSchedule2023, mockSchedule2024]
        mockGetPlayoffBracket.side_effect = [
            mockPlayoffSchedule_2022,
            mockPlayoffSchedule_2023,
            mockPlayoffSchedule_2024,
        ]

        leagueLoader = MyFantasyLeagueLeagueLoader(
            "789",
            [2022, 2023, 2024],
            mflUsername="mflu",
            mflPassword="mflp",
            mflUserAgentName="mfluan",
            leagueName="custom name",
        )
        league = leagueLoader.loadLeague()

        self.assertEqual("custom name", league.name)



from typing import Optional
from fleaflicker.api.LeagueInfoAPIClient import LeagueInfoAPIClient
from fleaflicker.api.ScoringAPIClient import ScoringAPIClient
from fleaflicker.enum.Sport import Sport
from sleeper.enum import Sport

from leeger.enum.MatchupType import MatchupType
from leeger.league_loader.LeagueLoader import LeagueLoader
from leeger.model.league.League import League
from leeger.model.league.Matchup import Matchup
from leeger.model.league.Owner import Owner
from leeger.model.league.Team import Team
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year
from leeger.validate import leagueValidation
from leeger.model.league.Division import Division


class FleaflickerLeagueLoader(LeagueLoader):
    """
    Responsible for loading a League from Fleaflicker.
    https://www.fleaflicker.com/
    """

    def __init__(
        self,
        leagueId: str,
        years: list[int],
        *,
        ownerNamesAndAliases: Optional[dict[str, list[str]]] = None,
        leagueName: Optional[str] = None,
    ):
        # validation
        try:
            int(leagueId)
        except ValueError:
            raise ValueError(f"League ID '{leagueId}' could not be turned into an int.")
        super().__init__(
            leagueId, years, ownerNamesAndAliases=ownerNamesAndAliases, leagueName=leagueName
        )

        self.__fleaflickerTeamIdToOwnerMap: dict[int, Owner] = dict()
        self.__fleaflickerTeamIdToTeamMap: dict[int, Team] = dict()
        self.__fleaflickerDivisionIdToDivisionMap: dict[
            int, Division
        ] = dict()  # holds the division info for ONLY the current year

    def __getAllLeagues(self) -> list[dict]:
        # return a list of all leagues
        fleaflickerLeagues = list()
        for year in self._years:
            fleaflickerLeagues.append(
                LeagueInfoAPIClient.get_league_standings(
                    sport=Sport.NFL, league_id=int(self._leagueId), season=year
                )
            )
        self._validateRetrievedLeagues(fleaflickerLeagues)
        return fleaflickerLeagues

    def getOwnerNames(self) -> dict[int, list[str]]:
        yearToOwnerNamesMap: dict[int, list[str]] = dict()
        fleaflickerLeagues = self.__getAllLeagues()
        for fleaflickerLeague in fleaflickerLeagues:
            yearToOwnerNamesMap[int(fleaflickerLeague["season"])] = list()
            for division in fleaflickerLeague["divisions"]:
                for team in division["teams"]:
                    ownerName = team["owners"][0]["displayName"]
                    yearToOwnerNamesMap[self._years[0]].append(ownerName)
        return yearToOwnerNamesMap

    def loadLeague(self, validate: bool = True) -> League:
        fleaflickerLeagues = self.__getAllLeagues()
        league = self.__buildLeague(fleaflickerLeagues)
        if validate:
            # validate new league
            leagueValidation.runAllChecks(league)
        self._warnForUnusedOwnerNames(league)
        return league

    def __buildLeague(self, fleaflickerLeagues: list[dict]) -> League:
        years = list()
        self.__loadOwners(fleaflickerLeagues)
        owners = list(self.__fleaflickerTeamIdToOwnerMap.values())
        for fleaflickerLeague in fleaflickerLeagues:
            # save league name for each year
            self._leagueNameByYear[fleaflickerLeague["season"]] = fleaflickerLeague["league"][
                "name"
            ]
            years.append(self.__buildYear(fleaflickerLeague))
        return League(name=self._getLeagueName(), owners=owners, years=self._getValidYears(years))

    def __buildYear(self, fleaflickerLeague: dict) -> Year:
        # save division info
        for fleaflickerDivision in fleaflickerLeague["divisions"]:
            self.__fleaflickerDivisionIdToDivisionMap[fleaflickerDivision["id"]] = Division(
                name=fleaflickerDivision["name"]
            )
        teams = self.__buildTeams(fleaflickerLeague)
        weeks = self.__buildWeeks(fleaflickerLeague)
        year = Year(
            yearNumber=int(fleaflickerLeague["season"]),
            teams=teams,
            weeks=weeks,
            divisions=list(self.__fleaflickerDivisionIdToDivisionMap.values()),
        )
        # clear division info
        self.__fleaflickerDivisionIdToDivisionMap = dict()
        return year

    def __buildWeeks(self, fleaflickerLeague: dict) -> list[Week]:
        weeks = list()
        # get all weeks
        fleaflicker_league_scoreboard = ScoringAPIClient.get_league_scoreboard(
            sport=Sport.NFL,
            league_id=fleaflickerLeague["league"]["id"],
            season=fleaflickerLeague["season"],
        )
        number_of_scoring_periods = (
            len(fleaflicker_league_scoreboard["eligibleSchedulePeriods"]) + 1
        )
        for scoring_period in range(1, number_of_scoring_periods):
            matchups = list()
            # get all games for this week
            current_scoreboard = ScoringAPIClient.get_league_scoreboard(
                sport=Sport.NFL,
                league_id=fleaflickerLeague["league"]["id"],
                season=fleaflickerLeague["season"],
                scoring_period=scoring_period,
            )
            for game in current_scoreboard.get("games", list()):
                # team A
                teamAFleaflicker: dict = game["away"]
                teamA = self.__fleaflickerTeamIdToTeamMap[teamAFleaflicker["id"]]
                teamAScore = game["awayScore"]["score"].get(
                    "value", 0
                )  # if "value" isn't found, score is 0

                # team B
                teamBFleaflicker: dict = game["home"]
                teamB = self.__fleaflickerTeamIdToTeamMap[teamBFleaflicker["id"]]
                teamBScore = game["homeScore"]["score"].get(
                    "value", 0
                )  # if "value" isn't found, score is 0

                # figure out tiebreakers
                teamAHasTieBreaker = game.get("awayResult") == "WIN"
                teamBHasTieBreaker = game.get("homeResult") == "WIN"

                # figure out matchup type
                matchupType = MatchupType.REGULAR_SEASON
                if (
                    game.get("isPlayoffs")
                    or game.get("isConsolation")
                    or game.get("isThirdPlaceGame")
                ):
                    matchupType = MatchupType.PLAYOFF
                if game.get("isChampionshipGame"):
                    matchupType = MatchupType.CHAMPIONSHIP

                # only add matchup if it is completed
                if game.get("isFinalScore"):
                    matchups.append(
                        Matchup(
                            teamAId=teamA.id,
                            teamBId=teamB.id,
                            teamAScore=teamAScore,
                            teamBScore=teamBScore,
                            teamAHasTiebreaker=teamAHasTieBreaker,
                            teamBHasTiebreaker=teamBHasTieBreaker,
                            matchupType=matchupType,
                        )
                    )
            if len(matchups) > 0:
                weeks.append(Week(weekNumber=scoring_period, matchups=matchups))
        return weeks

    def __buildTeams(self, fleaflickerLeague: dict) -> list[Team]:
        teams = list()
        for division in fleaflickerLeague["divisions"]:
            for team in division["teams"]:
                teamName = team["name"]
                teamId = team["id"]
                owner = self.__fleaflickerTeamIdToOwnerMap[teamId]
                team = Team(
                    ownerId=owner.id,
                    name=teamName,
                    divisionId=self.__fleaflickerDivisionIdToDivisionMap[division["id"]].id,
                )
                teams.append(team)
                self.__fleaflickerTeamIdToTeamMap[teamId] = team
        return teams

    def __loadOwners(self, fleaflickerLeagues: list[dict]) -> None:
        for fleaflickerLeague in fleaflickerLeagues:
            for division in fleaflickerLeague["divisions"]:
                for team in division["teams"]:
                    if "owners" not in team:
                        # some teams don't have owners listed in them, use the team name instead
                        ownerName = team["name"]
                    else:
                        ownerName = team["owners"][0]["displayName"]
                    # get general owner name if there is one
                    generalOwnerName = self._getGeneralOwnerNameFromGivenOwnerName(ownerName)
                    ownerName = generalOwnerName if generalOwnerName is not None else ownerName
                    self.__fleaflickerTeamIdToOwnerMap[team["id"]] = Owner(name=ownerName)



from __future__ import annotations

from dataclasses import dataclass
from leeger.model.abstract.EqualityCheck import EqualityCheck

from leeger.model.abstract.UniqueId import UniqueId
from leeger.util.CustomLogger import CustomLogger
from leeger.util.JSONDeserializable import JSONDeserializable
from leeger.util.JSONSerializable import JSONSerializable
from leeger.util.equality import modelEquals


@dataclass(kw_only=True, eq=False)
class Division(UniqueId, EqualityCheck, JSONSerializable, JSONDeserializable):
    __LOGGER = CustomLogger.getLogger()
    name: str

    def equals(
        self,
        otherDivision: Division,
        *,
        ignoreIds: bool = False,
        ignoreBaseIds: bool = False,
        logDifferences: bool = False,
    ) -> bool:
        """
        Checks if *this* Division is the same as the given Division.
        """

        return modelEquals(
            objA=self,
            objB=otherDivision,
            baseFields={"name"},
            parentKey="Division",
            ignoreIdFields=ignoreIds,
            ignoreBaseIdField=ignoreBaseIds,
            logDifferences=logDifferences,
        )

    def __eq__(self, otherDivision: Division) -> bool:
        self.__LOGGER.info("Use .equals() for more options when comparing Division instances.")
        return self.equals(otherDivision=otherDivision)

    def toJson(self) -> dict:
        return {"id": self.id, "name": self.name}

    @staticmethod
    def fromJson(d: dict) -> Division:
        division = Division(name=d["name"])
        division.id = d["id"]
        return division



from typing import Any, Optional

from leeger.util.CustomLogger import CustomLogger
from leeger.util.GeneralUtil import GeneralUtil


def modelEquals(
    *,
    objA: Any,
    objB: Any,
    baseFields: set[str],
    idFields: Optional[set[str]] = None,
    parentKey: str = "",
    ignoreIdFields: bool = False,
    ignoreBaseIdField: bool = False,
    logDifferences: bool = False,
    toJsonMethodName: str = "toJson",
    equalityFunctionMap: Optional[dict[str, callable]] = None,
    equalityFunctionKwargsMap: Optional[dict[str, dict]] = None,
) -> bool:
    """
    Checks if objA is the same as objB.

    objA: instance to compare
    objB: instance to compare
    baseFields: non-id fields
    idFields: id fields
    parentKey: name of the parent key to display in the differences log
    ignoreIds: whether to compare id fields or not (does not include 'id' for any model)
    ignoreBaseIdField: whether to compare base id fields (id) or not
    logDifferences: whether to log differences in the case of inequality or not
    toJsonMethodName: the name of the method to call on these objects that will return the JSON representation of it
    equalityFunctionMap: maps the names of fields to a custom equality function that should be called for them*
    equalityFunctionKwargsMap: maps the names of fields to the kwargs that should be passed into their equalityFunctionMap

    *
    - equality functions should take at least 2 parameters (1 parameter for each field value).
    - equality functions should return a boolean.
    - equality functions should take **kwargs as the 3rd parameter.
    - equality functions will be passed the kwargs from equalityFunctionKwargsMap that match the field name (if given)
    """
    LOGGER = CustomLogger.getLogger()

    equalityFunctionMap = dict() if equalityFunctionMap is None else equalityFunctionMap
    equalityFunctionKwargsMap = (
        dict() if equalityFunctionKwargsMap is None else equalityFunctionKwargsMap
    )

    def isEqual(field: str, objA: Any, objB: Any) -> bool:
        if field in equalityFunctionMap:
            equalityFunction = equalityFunctionMap[field]
            # see if we should pass any kwargs
            kwargsToPass = (
                dict()
                if field not in equalityFunctionKwargsMap
                else equalityFunctionKwargsMap[field]
            )
            return equalityFunction(getattr(objA, field), getattr(objB, field), **kwargsToPass)
        return getattr(objA, field) == getattr(objB, field)

    equal = True
    for field in baseFields:
        equal = equal and isEqual(field, objA, objB)

    if not ignoreIdFields and idFields:
        for field in idFields:
            equal = equal and isEqual(field, objA, objB)

    if not ignoreBaseIdField:
        if hasattr(objA, "id") and hasattr(objB, "id"):
            equal = equal and isEqual("id", objA, objB)
        # weird case: one instance has an id field and the other doesn't
        if hasattr(objA, "id") != hasattr(objB, "id"):
            equal = False

    if not equal and logDifferences:
        ignoreKeyNames = list()
        if ignoreIdFields:
            ignoreKeyNames += idFields
        if ignoreBaseIdField:
            ignoreKeyNames.append("id")
        objAJson = getattr(objA, toJsonMethodName)()
        objBJson = getattr(objB, toJsonMethodName)()
        differences = GeneralUtil.findDifferentFields(
            objAJson, objBJson, parentKey=parentKey, ignoreKeyNames=ignoreKeyNames
        )
        LOGGER.info(f"Differences: {differences}")
    return equal



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


@dataclass
class EqualityCheck(ABC):
    """
    Model classes should inherit this in order to have a .equals() method.
    """

    @abstractmethod
    def equals(
        self,
        otherInstance: Any,
        *,
        ignoreIds: bool = False,
        ignoreBaseIds: bool = False,
        logDifferences: bool = False,
    ) -> bool:
        ...



import unittest
from leeger.exception.DoesNotExistException import DoesNotExistException
from leeger.exception.LeagueLoaderException import LeagueLoaderException

from leeger.league_loader.LeagueLoader import LeagueLoader
from leeger.model.league.League import League
from leeger.model.league.Owner import Owner
from leeger.model.league.Week import Week
from leeger.model.league.Year import Year


class TestLeagueLoader(unittest.TestCase):
    def test_noYearsGiven(self):
        with self.assertRaises(ValueError) as context:
            LeagueLoader("0", [])
        self.assertEqual("No years given to load league with ID '0'.", str(context.exception))

    def test_yearsGivenThatArentInts(self):
        with self.assertRaises(ValueError) as context:
            LeagueLoader("0", ["a"])
        self.assertEqual("All given years must be ints.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            LeagueLoader("0", ["1"])
        self.assertEqual("All given years must be ints.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            LeagueLoader("0", [1, "1"])
        self.assertEqual("All given years must be ints.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            LeagueLoader("0", ["1", 1])
        self.assertEqual("All given years must be ints.", str(context.exception))

        # test for no error
        LeagueLoader("0", [1])

    def test__getGeneralOwnerNameFromGivenOwnerName(self):
        leagueLoader = LeagueLoader(
            "leagueId",
            [2021],
            ownerNamesAndAliases={
                "John Smith": ["John", "Smith"],
                "Jane Doe": ["Jane", "Doe"],
                "Tom Hanks": ["Tom"],
            },
        )

        # alias that belongs to a single owner
        self.assertEqual("John Smith", leagueLoader._getGeneralOwnerNameFromGivenOwnerName("John"))

        # alias that belongs to multiple owners
        self.assertEqual("John Smith", leagueLoader._getGeneralOwnerNameFromGivenOwnerName("Smith"))

        # alias that does not belong to any owner
        self.assertIsNone(leagueLoader._getGeneralOwnerNameFromGivenOwnerName("Foo"))

        # owner name that is also an owner name but NOT an alias
        self.assertIsNone(leagueLoader._getGeneralOwnerNameFromGivenOwnerName("John Smith"))

        # owner name that is not in the aliases list
        self.assertIsNone(leagueLoader._getGeneralOwnerNameFromGivenOwnerName("Foo"))

    def test__getOwnerByName(self):
        owners = [Owner(name="John Smith"), Owner(name="Jane Doe"), Owner(name="Tom Hanks")]
        leagueLoader = LeagueLoader(
            "leagueId",
            [2021],
            ownerNamesAndAliases={
                "John Smith": ["John", "Smith"],
                "Jane Doe": ["Jane", "Doe"],
                "Tom Hanks": ["Tom"],
            },
        )
        leagueLoader._owners = owners

        # exact owner name match
        self.assertEqual(owners[0], leagueLoader._getOwnerByName("John Smith"))

        # aliased owner name match
        self.assertEqual(owners[0], leagueLoader._getOwnerByName("John"))

        # alias that belongs to a single owner
        self.assertEqual(owners[0], leagueLoader._getOwnerByName("Smith"))

        # alias that does not exist
        with self.assertRaises(DoesNotExistException):
            leagueLoader._getOwnerByName("Foo")

    def test__getLeagueName(self):
        # no league name given, takes most recent year name
        leagueLoader = LeagueLoader("leagueId", [2021])
        leagueLoader._leagueNameByYear = {2022: "foo", 2019: "baz", 2021: "bar"}
        leagueName = leagueLoader._getLeagueName()
        self.assertEqual("foo", leagueName)

        # league name given, league names by year set, takes league name given
        leagueLoader = LeagueLoader("leagueId", [2021], leagueName="bot")
        leagueLoader._leagueNameByYear = {2022: "foo", 2019: "baz", 2021: "bar"}
        leagueName = leagueLoader._getLeagueName()
        self.assertEqual("bot", leagueName)

        # league name given, no league names by year set, takes league name given
        leagueLoader = LeagueLoader("leagueId", [2021], leagueName="bot")
        leagueName = leagueLoader._getLeagueName()
        self.assertEqual("bot", leagueName)

        # no league name given, no league names by year set, raises exception
        leagueLoader = LeagueLoader("leagueId", [2021])

        with self.assertRaises(LeagueLoaderException) as context:
            leagueLoader._getLeagueName()
        self.assertEqual(
            "Tried to retrieve league name with no leagueName parameter given and no league names set.",
            str(context.exception),
        )

    def test__getValidYears(self):
        # no change -> returns same years
        dummyWeeks = [Week(weekNumber=1, matchups=list())]
        year2020 = Year(yearNumber=2020, teams=list(), weeks=dummyWeeks)
        year2021 = Year(yearNumber=2021, teams=list(), weeks=dummyWeeks)
        year2022 = Year(yearNumber=2022, teams=list(), weeks=dummyWeeks)
        leagueLoader = LeagueLoader("leagueId", [2020, 2021, 2022])
        response = leagueLoader._getValidYears([year2020, year2021, year2022])
        self.assertEqual([year2020, year2021, year2022], response)

        # year with no weeks -> removes year
        dummyWeeks = [Week(weekNumber=1, matchups=list())]
        year2020 = Year(yearNumber=2020, teams=list(), weeks=dummyWeeks)
        year2021 = Year(yearNumber=2021, teams=list(), weeks=list())
        year2022 = Year(yearNumber=2022, teams=list(), weeks=dummyWeeks)
        leagueLoader = LeagueLoader("leagueId", [2020, 2021, 2022])
        response = leagueLoader._getValidYears([year2020, year2021, year2022])
        self.assertEqual([year2020, year2022], response)

        # years not sorted correctly -> sorts years
        dummyWeeks = [Week(weekNumber=1, matchups=list())]
        year2020 = Year(yearNumber=2020, teams=list(), weeks=dummyWeeks)
        year2021 = Year(yearNumber=2021, teams=list(), weeks=dummyWeeks)
        year2022 = Year(yearNumber=2022, teams=list(), weeks=dummyWeeks)
        leagueLoader = LeagueLoader("leagueId", [2020, 2021, 2022])
        response = leagueLoader._getValidYears([year2021, year2022, year2020])
        self.assertEqual([year2020, year2021, year2022], response)

    def test__warnForUnusedOwnerNames(self):
        # ownerNamesAndAliases not given, nothing is logged
        owner1 = Owner(name="o1")
        owner2 = Owner(name="o2")
        league = League(name="league", owners=[owner1, owner2], years=list())
        leagueLoader = LeagueLoader("leagueId", [2020])

        try:
            with self.assertLogs() as _:
                leagueLoader._warnForUnusedOwnerNames(league)
        except AssertionError:
            pass
        else:
            self.fail("Logs were produced but weren't expected")

        # ownerNamesAndAliases given, all are used, nothing is logged
        leagueLoader = LeagueLoader(
            "leagueId", [2020], ownerNamesAndAliases={"o1": list(), "o2": list()}
        )
        try:
            with self.assertLogs() as _:
                leagueLoader._warnForUnusedOwnerNames(league)
        except AssertionError:
            pass
        else:
            self.fail("Logs were produced but weren't expected")

        # ownerNamesAndAliases given, some aren't used, logs each unused owner name
        leagueLoader = LeagueLoader(
            "leagueId",
            [2020],
            ownerNamesAndAliases={"o1": list(), "o2": list(), "o3": list(), "o4": list()},
        )

        with self.assertLogs() as captured:
            leagueLoader._warnForUnusedOwnerNames(league)
        self.assertEqual(
            "Some owner names were given but not assigned to the loaded League: ['o3', 'o4']",
            str(captured.records[0].getMessage()),
        )



from dataclasses import dataclass
import unittest

from typing import Any, Optional

from leeger.util.equality import modelEquals
from unittest.mock import patch

# helper stuff


@dataclass
class Foo:
    field1: str
    field2: int
    idField1: str
    nestedField: Optional[Any] = None
    id: str = ""

    def fooEquals(self, otherFoo: Any, *, ignoreIdFields: bool = False) -> bool:
        return modelEquals(
            objA=self,
            objB=otherFoo,
            baseFields={"field1", "field2", "nestedField"},
            idFields={"idField1"},
            parentKey="Foo",
            ignoreIdFields=ignoreIdFields,
        )

    def toJson(self):
        return {"field1": self.field1, "field2": self.field2, "idField1": self.idField1}


class TestEquality(unittest.TestCase):
    @patch("leeger.util.CustomLogger.CustomLogger.getLogger")
    def test_equals(self, mockGetLogger):
        mockLogger = mockGetLogger.return_value

        # basic equal object
        objA = Foo("a", 1, "id")
        objB = Foo("a", 1, "id")

        result = modelEquals(
            objA=objA, objB=objB, baseFields={"field1", "field2"}, idFields={"idField1"}
        )

        self.assertTrue(result)
        mockLogger.info.assert_not_called()

        # basic unequal object
        objA = Foo("a", 1, "id")
        objB = Foo("b", 1, "id")

        result = modelEquals(
            objA=objA, objB=objB, baseFields={"field1", "field2"}, idFields={"idField1"}
        )

        self.assertFalse(result)
        mockLogger.info.assert_not_called()

        # basic equal object when ignoring id fields
        objA = Foo("a", 1, "id")
        objB = Foo("a", 1, "id2")

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            ignoreIdFields=True,
        )

        self.assertTrue(result)
        mockLogger.info.assert_not_called()

        # basic unequal object when not ignoring id fields
        objA = Foo("a", 1, "id")
        objB = Foo("a", 1, "id2")

        result = modelEquals(
            objA=objA, objB=objB, baseFields={"field1", "field2"}, idFields={"idField1"}
        )

        self.assertFalse(result)
        mockLogger.info.assert_not_called()

        # basic equal object when ignoring  base id field
        objA = Foo("a", 1, "id", id="id1")
        objB = Foo("a", 1, "id", id="id1")
        mockLogger.info.assert_not_called()

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            ignoreIdFields=True,
        )

        self.assertTrue(result)

        # basic unequal object when not ignoring base id field
        objA = Foo("a", 1, "id", id="id1")
        objB = Foo("a", 1, "id", id="id2")

        result = modelEquals(
            objA=objA, objB=objB, baseFields={"field1", "field2"}, idFields={"idField1"}
        )

        self.assertFalse(result)
        mockLogger.info.assert_not_called()

        # basic unequal object logging differences
        objA = Foo("a", 1, "id")
        objB = Foo("b", 1, "id")

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            logDifferences=True,
        )

        self.assertFalse(result)
        mockLogger.info.assert_called_once_with("Differences: [('field1', ('a', 'b'))]")
        mockLogger.reset_mock()

        # basic unequal object logging differences with parent key
        objA = Foo("a", 1, "id")
        objB = Foo("b", 1, "id")

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            logDifferences=True,
            parentKey="Foo",
        )

        self.assertFalse(result)
        mockLogger.info.assert_called_once_with("Differences: [('Foo.field1', ('a', 'b'))]")
        mockLogger.reset_mock()

        # basic unequal object logging differences with json method name
        objA = Foo("a", 1, "id")
        objB = Foo("b", 1, "id")

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            logDifferences=True,
            toJsonMethodName="toJson",
        )

        self.assertFalse(result)
        mockLogger.info.assert_called_once_with("Differences: [('field1', ('a', 'b'))]")
        mockLogger.reset_mock()

        # basic equal object with equality function map
        objA = Foo("a", 1, "id")
        objB = Foo("a", 1, "id")

        def checkField1(a, b, **kwargs):
            return a == b

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            equalityFunctionMap={"field1": checkField1},
        )

        self.assertTrue(result)
        mockLogger.info.assert_not_called()

        # basic equal object with equality function map with function that makes them unequal
        objA = Foo("a", 1, "id")
        objB = Foo("a", 1, "id")

        def checkField1(a, b, **kwargs):
            return False

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2"},
            idFields={"idField1"},
            equalityFunctionMap={"field1": checkField1},
        )

        self.assertFalse(result)
        mockLogger.info.assert_not_called()

        # basic equal object with nested field
        nestedFooA = Foo("b", 2, "id")
        nestedFooB = Foo("b", 2, "id")
        objA = Foo("a", 1, "id", [nestedFooA])
        objB = Foo("a", 1, "id", [nestedFooB])

        def checkNestedField(a, b, **kwargs):
            equals = True
            for curA, curB in zip(a, b):
                equals = equals and curA.fooEquals(curB, **kwargs)
            return equals

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2", "nestedField"},
            idFields={"idField1"},
            equalityFunctionMap={"nestedField": checkNestedField},
        )

        self.assertTrue(result)
        mockLogger.info.assert_not_called()

        # basic unequal object with nested field
        nestedFooA = Foo("b", 2, "idwrong")
        nestedFooB = Foo("b", 2, "id")
        objA = Foo("a", 1, "id", [nestedFooA])
        objB = Foo("a", 1, "id", [nestedFooB])

        def checkNestedField(a, b, **kwargs):
            equals = True
            for curA, curB in zip(a, b):
                equals = equals and curA.fooEquals(curB, **kwargs)
            return equals

        result = modelEquals(
            objA=objA,
            objB=objB,
            baseFields={"field1", "field2", "nestedField"},
            idFields={"idField1"},
            equalityFunctionMap={"nestedField": checkNestedField},
        )

        self.assertFalse(result)
        mockLogger.info.assert_not_called()



import unittest
from leeger.model.league.Division import Division


class TestDivision(unittest.TestCase):
    def test_division_init(self):
        division = Division(name="name")

        self.assertEqual("name", division.name)

    def test_division_eq_callsEqualsMethod(self):
        # create division 1
        division_1 = Division(name="division")

        # create division 2
        division_2 = Division(name="division")
        division_2.id = division_1.id

        result = division_1 == division_2

        self.assertTrue(result)

    def test_division_eq_equal(self):
        # create division 1
        division_1 = Division(name="division")

        # create division 2
        division_2 = Division(name="division")

        result = division_1.equals(division_2, ignoreBaseIds=True)

        self.assertTrue(result)

    def test_division_eq_notEqual(self):
        # create division 1
        division_1 = Division(name="division")

        # create division 2
        division_2 = Division(name="divisionDIF")

        result = division_1.equals(division_2, ignoreBaseIds=True)

        self.assertFalse(result)

    def test_toFromJson(self):
        division = Division(name="division")
        self.assertEqual(division, Division.fromJson(division.toJson()))
